/*
 * SPDX-FileCopyrightText: 2023 Espressif Systems (Shanghai) CO LTD
 *
 * SPDX-License-Identifier: CC0-1.0
 */

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "driver/gpio.h"
#include "esp_lcd_panel_io.h"
#include "esp_lcd_panel_ops.h"
#include "esp_lcd_panel_rgb.h"
#include "esp_log.h"

#include "sdkconfig.h"
#include "../priv_include/bsp_err_check.h"
#include "../priv_include/bsp_sub_board.h"
#include "bsp/esp-bsp.h"
#include "esp_io_expander.h"

static const char *TAG = "bsp_lcd";
#if BSP_IO_EXPANDER_EN
static esp_io_expander_handle_t io = NULL;
#endif
static bsp_lcd_trans_done_cb_t trans_done = NULL;
static esp_err_t lcd_config(void);

/**************************************************************************************************
 *
 * LCD Panel Function
 *
 **************************************************************************************************/
IRAM_ATTR static bool on_vsync_event(esp_lcd_panel_handle_t panel, esp_lcd_rgb_panel_event_data_t *edata, void *user_ctx)
{
    BaseType_t need_yield = pdFALSE;
    if (trans_done) {
        if (trans_done(panel)) {
            need_yield = pdTRUE;
        }
    }

    return (need_yield == pdTRUE);
}

esp_lcd_panel_handle_t bsp_lcd_init(void *arg)
{

#if BSP_IO_EXPANDER_EN
    BSP_NULL_CHECK(arg, ESP_ERR_INVALID_ARG);
    io = (esp_io_expander_handle_t)arg;
#endif
    BSP_ERROR_CHECK_RETURN_ERR(lcd_config());

    ESP_LOGI(TAG, "Initialize RGB panel");
    esp_lcd_panel_handle_t panel_handle = NULL;
    esp_lcd_rgb_panel_config_t panel_conf = {
        .clk_src = LCD_CLK_SRC_PLL160M,
        .psram_trans_align = 64,
        .data_width = 16,
        .de_gpio_num = BSP_LCD_DE,
        .pclk_gpio_num = BSP_LCD_PCLK,
        .vsync_gpio_num = BSP_LCD_VSYNC,
        .hsync_gpio_num = BSP_LCD_HSYNC,
        .data_gpio_nums = {
            BSP_LCD_DATA0,
            BSP_LCD_DATA1,
            BSP_LCD_DATA2,
            BSP_LCD_DATA3,
            BSP_LCD_DATA4,
            BSP_LCD_DATA5,
            BSP_LCD_DATA6,
            BSP_LCD_DATA7,
            BSP_LCD_DATA8,
            BSP_LCD_DATA9,
            BSP_LCD_DATA10,
            BSP_LCD_DATA11,
            BSP_LCD_DATA12,
            BSP_LCD_DATA13,
            BSP_LCD_DATA14,
            BSP_LCD_DATA15,
        },
        .timings = {
            .pclk_hz = BSP_LCD_PIXEL_CLOCK_HZ,
            .h_res = BSP_LCD_H_RES,
            .v_res = BSP_LCD_V_RES,
            .hsync_back_porch = BSP_LCD_HSYNC_BACK_PORCH,
            .hsync_front_porch = BSP_LCD_HSYNC_FRONT_PORCH,
            .hsync_pulse_width = BSP_LCD_HSYNC_PULSE_WIDTH,
            .vsync_back_porch = BSP_LCD_VSYNC_BACK_PORCH,
            .vsync_front_porch = BSP_LCD_VSYNC_FRONT_PORCH,
            .vsync_pulse_width = BSP_LCD_VSYNC_PULSE_WIDTH,
            .flags.pclk_active_neg = BSP_LCD_PCLK_ACTIVE_NEG,
        },
        .flags.fb_in_psram = 1,
        .bounce_buffer_size_px = CONFIG_BSP_LCD_RGB_BOUNCE_BUFFER_HEIGHT * BSP_LCD_H_RES,
    };
    BSP_ERROR_CHECK_RETURN_NULL(esp_lcd_new_rgb_panel(&panel_conf, &panel_handle));
    esp_lcd_rgb_panel_event_callbacks_t cbs = {
        .on_vsync = on_vsync_event,
    };
    esp_lcd_rgb_panel_register_event_callbacks(panel_handle, &cbs, NULL);
    esp_lcd_panel_reset(panel_handle);
    esp_lcd_panel_init(panel_handle);

    return panel_handle;
}

esp_err_t bsp_lcd_register_trans_done_callback(bsp_lcd_trans_done_cb_t callback)
{
#if CONFIG_LCD_RGB_ISR_IRAM_SAFE
    if (callback) {
        ESP_RETURN_ON_FALSE(esp_ptr_in_iram(callback), ESP_ERR_INVALID_ARG, TAG, "Callback not in IRAM");
    }
#endif
    trans_done = callback;

    return ESP_OK;
}

/**************************************************************************************************
 *
 * LCD Configuration Function
 *
 **************************************************************************************************/
#define Delay(t)        vTaskDelay(pdMS_TO_TICKS(t))
#define udelay(t)       esp_rom_delay_us(t)
#if BSP_IO_EXPANDER_EN
#define CS(n)           BSP_ERROR_CHECK_RETURN_ERR(esp_io_expander_set_level(io, BSP_LCD_SPI_CS, n))
#define SCK(n)          BSP_ERROR_CHECK_RETURN_ERR(esp_io_expander_set_level(io, BSP_LCD_SPI_SCK, n))
#define SDO(n)          BSP_ERROR_CHECK_RETURN_ERR(esp_io_expander_set_level(io, BSP_LCD_SPI_SDO, n))
#define RST(n)          BSP_ERROR_CHECK_RETURN_ERR(esp_io_expander_set_level(io, BSP_LCD_RST, n))
#else
#define CS(n)           BSP_ERROR_CHECK_RETURN_ERR(gpio_set_level(BSP_LCD_SPI_CS, n))
#define SCK(n)          BSP_ERROR_CHECK_RETURN_ERR(gpio_set_level(BSP_LCD_SPI_SCK, n))
#define SDO(n)          BSP_ERROR_CHECK_RETURN_ERR(gpio_set_level(BSP_LCD_SPI_SDO, n))
#define RST(n)          BSP_ERROR_CHECK_RETURN_ERR(gpio_set_level(BSP_LCD_RST, n))
#endif

/**
 * @brief Simulate SPI to write data using io expander
 *
 * @param data: Data to write
 *
 * @return
 *      - ESP_OK: Success, otherwise returns ESP_ERR_xxx
 */
static esp_err_t spi_write(uint16_t data)
{
    for (uint8_t n = 0; n < 9; n++) {
        if (data & 0x0100) {
            SDO(1);
        } else {
            SDO(0);
        }
        data = data << 1;

        SCK(0);
        udelay(10);
        SCK(1);
        udelay(10);
    }

    return ESP_OK;
}

/**
 * @brief Simulate SPI to write LCD command using io expander
 *
 * @param data: LCD command to write
 *
 * @return
 *      - ESP_OK: Success, otherwise returns ESP_ERR_xxx
 */
static esp_err_t spi_write_cmd(uint16_t data)
{
    CS(0);
    udelay(10);

    spi_write((data & 0x00FF));

    udelay(10);
    CS(1);
    SCK(0);
    SDO(0);
    udelay(10);

    return ESP_OK;
}

/**
 * @brief Simulate SPI to write LCD data using io expander
 *
 * @param data: LCD data to write
 *
 * @return
 *      - ESP_OK: Success, otherwise returns ESP_ERR_xxx
 */
static esp_err_t spi_write_data(uint16_t data)
{
    CS(0);
    udelay(10);

    data &= 0x00FF;
    data |= 0x0100;
    spi_write(data);

    udelay(10);
    CS(1);
    SCK(0);
    SDO(0);
    udelay(10);

    return ESP_OK;
}

#define LCD_CONFIG_DATA_LEN_MAX         (16)

/**
 * @brief LCD configuration data structure type
 *
 */
typedef struct {
    uint8_t cmd;            // LCD command
    uint8_t data[LCD_CONFIG_DATA_LEN_MAX];       // LCD data
    uint8_t data_bytes;     // Length of data in above data array; 0xFF = end of cmds.
} lcd_config_data_t;

// *INDENT-OFF*
const static lcd_config_data_t LCD_CONFIG_CMD_1[] = {
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x13}, 5},
    {0xef, {0x08}, 1},
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x10}, 5},
    {0xc0, {0x3b, 0x00}, 2},
    {0xc1, {0x0b, 0x02}, 2},
    {0xc2, {0x37, 0x02}, 2},
    {0xcc, {0x10}, 1},
    {0xb0, {0x00, 0x1d, 0x29, 0x12, 0x17, 0x0b, 0x18, 0x09, 0x08, 0x2a, 0x07, 0x14, 0x11, 0x27, 0x32, 0x1f}, 16},
    {0xb1, {0x00, 0x1d, 0x29, 0x12, 0x16, 0x0a, 0x18, 0x08, 0x09, 0x2a, 0x07, 0x13, 0x12, 0x27, 0x33, 0x1f}, 16},
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x11}, 5},
    {0xb0, {0x6d}, 1},
    {0xb1, {0x24}, 1},
    {0xb2, {0x81}, 1},
    {0xb3, {0x80}, 1},
    {0xb5, {0x43}, 1},
    {0xb7, {0x85}, 1},
    {0xb8, {0x20}, 1},
    {0xc1, {0x78}, 1},
    {0xc2, {0x78}, 1},
    {0xd0, {0x88}, 1},
    {0xe0, {0x00, 0x00, 0x02}, 3},
    {0xe1, {0x03, 0xa0, 0x00, 0x00, 0x04, 0xa0, 0x00, 0x00, 0x00, 0x20, 0x20}, 11},
    {0xe2, {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 13},
    {0xe3, {0x00, 0x00, 0x11, 0x00}, 4},
    {0xe4, {0x22, 0x00}, 2},
    {0xe5, {0x05, 0xec, 0xa0, 0xa0, 0x07, 0xee, 0xa0, 0xa0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 16},
    {0xe6, {0x00, 0x00, 0x11, 0x00}, 4},
    {0xe7, {0x22, 0x00}, 2},
    {0xe8, {0x06, 0xed, 0xa0, 0xa0, 0x08, 0xef, 0xa0, 0xa0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 16},
    {0xeb, {0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00}, 7},
    {0xed, {0xff, 0xff, 0xff, 0xba, 0x0a, 0xbf, 0x45, 0xff, 0xff, 0x54, 0xfb, 0xa0, 0xab, 0xff, 0xff, 0xff}, 16},
    {0xef, {0x10, 0x0d, 0x04, 0x08, 0x3f, 0x1f}, 6},
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x13}, 5},
    {0xe8, {0x00, 0x0e}, 2},
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x00}, 5},
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x13}, 5},
    {0xe8, {0x00, 0x0c}, 2},
    {0xe8, {0x00, 0x00}, 2},
    {0xff, {0x77, 0x01, 0x00, 0x00, 0x00}, 5},
    // {0x3a, {0x66}, 1},      /* RGB666 */
    {0x3a, {0x55}, 1},      /* RGB565 */
    {0xcd, {0x00}, 1},
    {0x29, {0x00}, 0},

    {0x00, {0x00}, 0xff},
};
// *INDENT-OFF*

/**
 * @brief Configure LCD with specific commands and data
 *
 * @return
 *      - ESP_OK: Success, otherwise returns ESP_ERR_xxx
 *
 */
static esp_err_t lcd_config(void)
{
#if BSP_IO_EXPANDER_EN
    esp_io_expander_set_dir(io, BSP_LCD_SPI_CS, IO_EXPANDER_OUTPUT);
    esp_io_expander_set_dir(io, BSP_LCD_SPI_SCK, IO_EXPANDER_OUTPUT);
    esp_io_expander_set_dir(io, BSP_LCD_SPI_SDO, IO_EXPANDER_OUTPUT);
    if (BSP_LCD_RST != -1) {
        esp_io_expander_set_dir(io, BSP_LCD_RST, IO_EXPANDER_OUTPUT);
    }
    esp_io_expander_set_level(io, BSP_LCD_SPI_CS, 1);
    esp_io_expander_set_level(io, BSP_LCD_SPI_SCK, 1);
    esp_io_expander_set_level(io, BSP_LCD_SPI_SDO, 1);
#else
    uint64_t bit_msk = BIT64(BSP_LCD_SPI_CS) | BIT64(BSP_LCD_SPI_SCK) | BIT64(BSP_LCD_SPI_SDO);
    if (BSP_LCD_RST != -1) {
        bit_msk |= BIT64(BSP_LCD_RST);
    }
    gpio_config_t config = {
        .mode = GPIO_MODE_OUTPUT,
        .pull_down_en = GPIO_PULLDOWN_DISABLE,
        .pull_up_en = GPIO_PULLUP_DISABLE,
        .pin_bit_mask = bit_msk,
    };
    ESP_RETURN_ON_ERROR(gpio_config(&config), TAG, "SPI GPIO config failed");
#endif

    CS(1);
    SCK(1);
    SDO(1);
    if (BSP_LCD_RST != -1) {
        RST(0);
        Delay(10);
        RST(1);
        Delay(120);
    }

    BSP_ERROR_CHECK_RETURN_ERR(spi_write_cmd(0x11));
    vTaskDelay(pdMS_TO_TICKS(150));
    for (uint8_t i = 0; LCD_CONFIG_CMD_1[i].data_bytes != 0xff; i++) {
        BSP_ERROR_CHECK_RETURN_ERR(spi_write_cmd(LCD_CONFIG_CMD_1[i].cmd));
        for (uint8_t j = 0; j < LCD_CONFIG_CMD_1[i].data_bytes; j++) {
            BSP_ERROR_CHECK_RETURN_ERR(spi_write_data(LCD_CONFIG_CMD_1[i].data[j]));
        }
    }

#if !BSP_IO_EXPANDER_EN
    // gpio_reset_pin(BSP_LCD_SPI_CS);
    gpio_reset_pin(BSP_LCD_SPI_SCK);
    gpio_reset_pin(BSP_LCD_SPI_SDO);
#endif

    return ESP_OK;
}
